/*
 * Char type
 *
 * $Id: chr.c 1.3 Wed, 06 Jan 1999 02:35:46 +0100 crad $
 */
#include "s.h"

#ifdef COMMENT
/*
- les caracteres sont alloues dans un tableau qui n'est jamais collecte.
- on a uniquement des reference sur ce tableau.
*/
#endif

#define BSLASH	'\134'

struct CHR_SYM {
  char *str;
  char chr;
};

struct CHR_SYM csym[] = {
  { "null", 		0 },
  { "bell",			'\a' }, 
  { "backspace",	'\b' }, 
  { "tab",			'\t' }, 
  { "newline",		'\n' }, 
  { "vtab",			'\v' }, 
  { "formfeed",		'\f' }, 
  { "return",		'\r' },
  { "space",		' '  },
  { NULL}
};

Sobject scm_chr_array[256];

int scm_is_pointer_to_chr_array(SOBJ obj)
{
  return(obj >= scm_chr_array && obj < (scm_chr_array + 256));
}

static struct CHR_SYM *lookup_chr(int c)
{
  struct CHR_SYM *s = csym;
  while(s->str) {
	if (c == s->chr)	return(s);
	s++;
  }
  return(NULL);
}

static struct CHR_SYM *lookup_str(char *str)
{
  struct CHR_SYM *s = csym;
  while(s->str) {
	if (strcmp(str, s->str) == 0) return(s);
	s++;
  }
  return(NULL);
}


SOBJ scm_mkchar(char c)
{
  return(scm_chr_array + c);
}

void scm_char_print(SOBJ c, PORT *p)
{
  port_putc(p, SCM_CHAR(c));
}

void scm_char_write(SOBJ obj, PORT *p)
{
  int c = SCM_CHAR(obj);
  char buf[8];
  struct CHR_SYM *ps;

  port_puts(p, "#\134");
  if (c <= ' ' || c == 127) {
	ps = lookup_chr(c);
	if (ps != NULL) {
	  port_puts(p, ps->str);
	} else {
	  sprintf(buf, "%uo", c);  port_puts(p, buf);
	}
  } else {
	port_putc(p, c);
  }
}

int scm_char_reconize(PORT *port, int  c)
{
  return (c == '#' && port_peekc(port) == BSLASH);
}

SOBJ scm_char_parse(PORT *port, int c)
{
  struct CHR_SYM *ps;
  char str[80], *p;
  
  if (port_getc(port) != BSLASH) {
	fprintf(stderr, "scm_char_parse: internal error\n");
	exit(1);
  }
  c = port_getc(port);
  if (!isalnum(port_peekc(port))) {
	return(scm_mkchar(c));
  }
  str[0] = c; p = str+1;
  while(isalnum(port_peekc(port)) && p < (str + sizeof(str) - 1)) {
	*p++ = port_getc(port);
  }
  *p = 0;
  if (isdigit(*str)) {			/* octal number */
	c = strtol(str, NULL, 8);
  } else {
	ps = lookup_str(str);
	if (ps != NULL) {
	  c = ps->chr;
	} else {
	  fprintf(stderr, "#\\%s is not a char\n", str);
	  err("illegal char", NULL);
	}
  }
  return(scm_mkchar(c));
}

/*-- library */
SOBJ scm_charp(SOBJ obj)
{
  return(SCM_MKBOOL(SCM_CHARP(obj)));
}

static int ccmp(SOBJ x, SOBJ y)
{
  if (!SCM_CHARP(x)) 	err("charcmp: bad char", x);
  if (!SCM_CHARP(y)) 	err("charcmp: bad char", y);
  return(SCM_CHAR(x) - SCM_CHAR(y));
}

static int ccmpi(SOBJ x, SOBJ y)
{
  if (!SCM_CHARP(x)) 	err("charcmpi: bad char", x);
  if (!SCM_CHARP(y)) 	err("charcmpi: bad char", y);
  return(tolower(SCM_CHAR(x)) - tolower(SCM_CHAR(y)));
}

#define MKBOOL(x) ((x) ? scm_true: scm_false)

SOBJ scm_charlt(SOBJ x, SOBJ y) { return(SCM_MKBOOL(ccmp(x,y) <  0)); }
SOBJ scm_charle(SOBJ x, SOBJ y) { return(SCM_MKBOOL(ccmp(x,y) <= 0)); }
SOBJ scm_chareq(SOBJ x, SOBJ y) { return(SCM_MKBOOL(ccmp(x,y) == 0)); }
SOBJ scm_charge(SOBJ x, SOBJ y) { return(SCM_MKBOOL(ccmp(x,y) >= 0)); }
SOBJ scm_chargt(SOBJ x, SOBJ y) { return(SCM_MKBOOL(ccmp(x,y) >  0)); }

SOBJ scm_charlti(SOBJ x, SOBJ y) { return(SCM_MKBOOL(ccmpi(x,y) <  0)); }
SOBJ scm_charlei(SOBJ x, SOBJ y) { return(SCM_MKBOOL(ccmpi(x,y) <= 0)); }
SOBJ scm_chareqi(SOBJ x, SOBJ y) { return(SCM_MKBOOL(ccmpi(x,y) == 0)); }
SOBJ scm_chargei(SOBJ x, SOBJ y) { return(SCM_MKBOOL(ccmpi(x,y) >= 0)); }
SOBJ scm_chargti(SOBJ x, SOBJ y) { return(SCM_MKBOOL(ccmpi(x,y) >  0)); }

SOBJ scm_char_alphap(SOBJ x) 
{
  if (!SCM_CHARP(x)) 	err("char-alphabetic?: bad char", x);
  return(SCM_MKBOOL(isalpha(SCM_CHAR(x))));
}

SOBJ scm_char_nump(SOBJ x) 
{
  if (!SCM_CHARP(x)) 	err("char-numeric?: bad char", x);
  return(SCM_MKBOOL(isdigit(SCM_CHAR(x))));
}

SOBJ scm_char_whitep(SOBJ x) 
{
  if (!SCM_CHARP(x)) 	err("char-whitespace?: bad char", x);
  return(SCM_MKBOOL(isspace(SCM_CHAR(x))));
}

SOBJ scm_char_upperp(SOBJ x) 
{
  if (!SCM_CHARP(x)) 	err("char-upper-case?: bad char", x);
  return(SCM_MKBOOL(isupper(SCM_CHAR(x))));
}

SOBJ scm_char_lowerp(SOBJ x) 
{
  if (!SCM_CHARP(x)) 	err("char-lower-case?: bad char", x);
  return(SCM_MKBOOL(islower(SCM_CHAR(x))));
}

SOBJ scm_char2int(SOBJ x)
{
  if (!SCM_CHARP(x)) 	err("char->integer: bad char", x);
  return(SCM_MKINUM(SCM_CHAR(x)));
}

SOBJ scm_int2char(SOBJ x)
{
  int c = SCM_INUM(x);
  if (c < 0 || c >= 256) 	err("integer->char: bad integer", x);
  return(scm_mkchar(c));
}

SOBJ scm_charupc(SOBJ x)
{
  if (!SCM_CHARP(x))	err("char-upcase: bad char", x);
  return(scm_mkchar(toupper(SCM_CHAR(x))));
}

SOBJ scm_charlwc(SOBJ x)
{
  if (!SCM_CHARP(x))	err("char-downcase: bad char", x);
  return(scm_mkchar(tolower(SCM_CHAR(x))));
}

void scm_init_chr()
{
  int i;

  for (i = 0; i < 256; i++) {
	scm_chr_array[i].type = SOBJ_T_CHAR;
	SCM_CHAR(scm_chr_array + i) = i;
  }


  scm_add_cprim("char?",		scm_charp, 		1);
  scm_add_cprim("char=?",		scm_chareq, 	2);
  scm_add_cprim("char<?",		scm_charlt, 	2);
  scm_add_cprim("char>?",		scm_chargt, 	2);
  scm_add_cprim("char<=?",		scm_charle, 	2);
  scm_add_cprim("char>=?",		scm_charge, 	2);

  scm_add_cprim("char-ci=?",	scm_chareqi, 	2);
  scm_add_cprim("char-ci<?",	scm_charlti, 	2);
  scm_add_cprim("char-ci>?",	scm_chargti, 	2);
  scm_add_cprim("char-ci<=?",	scm_charlei, 	2);
  scm_add_cprim("char-ci>=?",	scm_chargei, 	2);

  scm_add_cprim("char-alphabetic?",		scm_char_alphap, 	1);
  scm_add_cprim("char-numeric?",		scm_char_nump,		1);
  scm_add_cprim("char-whitespace?",		scm_char_whitep,	1);
  scm_add_cprim("char-upper-case?",		scm_char_upperp,	1);
  scm_add_cprim("char-lower-case?",		scm_char_lowerp,	1);

  scm_add_cprim("char->integer",		scm_char2int,		1);
  scm_add_cprim("integer->char",		scm_int2char,		1);

  scm_add_cprim("char-upcase",		scm_charupc,	1);
  scm_add_cprim("char-downcase",	scm_charlwc,	1);

}

